!***********************************************************************************************************************************
!***********************************************************************************************************************************
!
!   ParaMonte: plain powerful parallel Monte Carlo library.
!
!   Copyright (C) 2012-present, The Computational Data Science Lab
!
!   This file is part of the ParaMonte library.
!
!   ParaMonte is free software: you can redistribute it and/or modify it
!   under the terms of the GNU Lesser General Public License as published
!   by the Free Software Foundation, version 3 of the License.
!
!   ParaMonte is distributed in the hope that it will be useful,
!   but WITHOUT ANY WARRANTY; without even the implied warranty of
!   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
!   GNU Lesser General Public License for more details.
!
!   You should have received a copy of the GNU Lesser General Public License
!   along with the ParaMonte library. If not, see,
!
!       https://github.com/cdslaborg/paramonte/blob/master/LICENSE
!
!   ACKNOWLEDGMENT
!
!   As per the ParaMonte library license agreement terms,
!   if you use any parts of this library for any purposes,
!   we ask you to acknowledge the use of the ParaMonte library
!   in your work (education/research/industry/development/...)
!   by citing the ParaMonte library as described on this page:
!
!       https://github.com/cdslaborg/paramonte/blob/master/ACKNOWLEDGMENT.md
!
!***********************************************************************************************************************************
!***********************************************************************************************************************************

module SpecMCMC_ChainSize_mod

    use Constants_mod, only: IK
    implicit none

    character(*), parameter         :: MODULE_NAME = "@SpecMCMC_ChainSize_mod"

    integer(IK)                     :: chainSize ! namelist input

    type                            :: ChainSize_type
        integer(IK)                 :: val
        integer(IK)                 :: def
        integer(IK)                 :: null
        character(:), allocatable   :: desc
    contains
        procedure, pass             :: set => setChainSize, checkForSanity, nullifyNameListVar
    end type ChainSize_type

    interface ChainSize_type
        module procedure            :: constructChainSize
    end interface ChainSize_type

    private :: constructChainSize, setChainSize, checkForSanity, nullifyNameListVar

!***********************************************************************************************************************************
!***********************************************************************************************************************************

contains

!***********************************************************************************************************************************
!***********************************************************************************************************************************

    function constructChainSize(methodName) result(ChainSizeObj)
#if defined DLL_ENABLED && !defined CFI_ENABLED
        !DEC$ ATTRIBUTES DLLEXPORT :: constructChainSize
#endif
        use Constants_mod, only: IK, NULL_IK
        use String_mod, only: num2str
        implicit none
        character(*), intent(in)    :: methodName
        type(ChainSize_type)        :: ChainSizeObj
        ChainSizeObj%def    = 100000_IK
        ChainSizeObj%null   = NULL_IK
        ChainSizeObj%desc   = &
        "chainSize determines the number of non-refined, potentially auto-correlated, but unique, samples &
        &drawn by the MCMC sampler before stopping " // methodName // ". &
        &For example, if you specify chainSize = 10000, then 10000 unique sample points (with no duplicates) will be drawn &
        &from the target objective function that the user has provided. &
        &The input value for chainSize must be a positive integer of a minimum value ndim+1 or larger, &
        &where ndim is the number of variables that define the domain of the objective function to be sampled. &
        &The default value is " // num2str(ChainSizeObj%def) // "."
    end function constructChainSize

!***********************************************************************************************************************************
!***********************************************************************************************************************************

    subroutine nullifyNameListVar(ChainSizeObj)
#if defined DLL_ENABLED && !defined CFI_ENABLED
        !DEC$ ATTRIBUTES DLLEXPORT :: nullifyNameListVar
#endif
        implicit none
        class(ChainSize_type), intent(in)  :: ChainSizeObj
        chainSize = ChainSizeObj%null
    end subroutine nullifyNameListVar

!***********************************************************************************************************************************
!***********************************************************************************************************************************

    subroutine setChainSize(ChainSizeObj,chainSize)
#if defined DLL_ENABLED && !defined CFI_ENABLED
        !DEC$ ATTRIBUTES DLLEXPORT :: setChainSize
#endif
        use Constants_mod, only: IK
        implicit none
        class(ChainSize_type), intent(inout)    :: ChainSizeObj
        integer(IK), intent(in)                 :: chainSize
        ChainSizeObj%val = chainSize
        if ( ChainSizeObj%val==ChainSizeObj%null ) then
            ChainSizeObj%val = ChainSizeObj%def
        end if
    end subroutine setChainSize

!***********************************************************************************************************************************
!***********************************************************************************************************************************

    subroutine checkForSanity(ChainSizeObj,Err,methodName,nd)
#if defined DLL_ENABLED && !defined CFI_ENABLED
        !DEC$ ATTRIBUTES DLLEXPORT :: checkForSanity
#endif
        use Constants_mod, only: IK
        use Err_mod, only: Err_type
        use String_mod, only: num2str
        implicit none
        class(ChainSize_type), intent(in)   :: ChainSizeObj
        integer(IK), intent(in)             :: nd
        character(*), intent(in)            :: methodName
        type(Err_type), intent(inout)       :: Err
        character(*), parameter             :: PROCEDURE_NAME = "@checkForSanity()"
        if ( ChainSizeObj%val<nd+1) then
            Err%occurred = .true.
            Err%msg =   Err%msg // &
                        MODULE_NAME // PROCEDURE_NAME // ": Error occurred. &
                        &The input requested value for chainSize (" // num2str(ChainSizeObj%val) // ") can neither be negative &
                        &nor smaller than ndim+1, where ndim represents the dimension of the sampling space, here ndim=" // &
                        num2str(nd) // ". &
                        &If you don't know an appropriate value for chainSize, drop it from the input list. " // &
                        methodName // " will automatically assign an appropriate value to it.\n\n"
        end if
    end subroutine checkForSanity

!***********************************************************************************************************************************
!***********************************************************************************************************************************

end module SpecMCMC_ChainSize_mod